const fs = require("fs");

const _ = require("lodash");
const moment = require("moment");
const watch = require("node-watch");
const shush = require("shush");
const async = require("async");
const argv = require("minimist")(process.argv.slice(2));

let config = shush("config.json");

// Loop through all scripts specified, and compile them, while building
// a list of dependencies to watch for changes and automatically recompile

// If a script is specified in console input, use that instead of config.json
if (argv._.length) {
  config.scripts = [
    {
      name: argv._[0],
      compiledName: argv._[0].replace(".txt", "") + "_compiled.txt"
    }
  ];
}

for (script of config.scripts) {
  recompile = (evt, name, script) => {
    // If a watcher is currently set, close it
    if (script.watcher) {
      script.watcher.close();
    }

    script.path = config.scriptsFolder + script.name;
    script.allDependents = [];

    fs.readFile(script.path, "utf-8", (err, data) => {
      if (err) {
        throw err;
      }
      script.contents = data;

      loopUntilNoDependents = script => {
        script.dependents = [];
        {
          let includeInstances = _.uniq(
            script.contents.match(/\w+\.txt(?=>)/g)
          );
          for (instance of includeInstances) {
            script.dependents.push({
              name: instance,
              path: config.scriptsFolder + instance
            });
          }
        }

        async.forEachOf(
          script.dependents,
          (dependent, key, callback) => {
            fs.readFile(dependent.path, "utf8", (err, data) => {
              if (err) return callback(err);
              dependent.contents = data;
              callback();
            });
          },
          err => {
            if (err) {
              console.error(err.message);
            }

            for (dependent of script.dependents) {
              script.contents = script.contents.replace(
                "$$<" + dependent.name + ">",
                dependent.contents
              );
            }

            // Fix corrupted $$< instances
            script.contents = script.contents.replace(/([^\$])\$</g, "$1$$$$<");

            // Add these dependencies to the full list
            script.allDependents = script.allDependents.concat(
              script.dependents
            );

            // If there's any remaining dependencies now, loop again.  Else, write our file
            // and set all watch files.
            if (/\w+\.txt(?=>)/.test(script.contents)) {
              loopUntilNoDependents(script);
            } else {
              // Remove whitespace if option is set
              if (
                config.removeWhitespace ||
                config.obfuscate ||
                config.blockFormat
              ) {
                // Remove all lines with "//" (comments)
                script.contents = script.contents.replace(/.*\/\/.*/gm, "");

                // Remove all empty lines and extraneous newlines
                script.contents = script.contents.replace(/^\s+/gm, "");
              }

              // If you are obfuscating, ensure you adhere to the these rules:
              // * no single-length string variable e.g. &j
              // * no non-variable instance of # or & in strings unless format code
              // * no operators in strings

              if (config.obfuscate || config.blockFormat) {
                // Get list of all local variables.  We skip flags (hard to parse)
                numericVariables = _.uniq(script.contents.match(/(?<!@)#\w+/g));
                stringVariables = _.uniq(script.contents.match(/(?<!@)&\w+/g));

                // Single length string variables are assumed to be formatting codes
                stringVariables = _.filter(stringVariables, a => {
                  return a.length > 2;
                });

                // Rebuild script from variables #q1, #q2, #q3, etc.

                var i = 1;
                for (variable of numericVariables) {
                  var regex = new RegExp(variable + "(?![a-z]|_)", "g");
                  script.contents = script.contents.replace(regex, "#q" + i);
                  i++;
                }

                // Do same for strings
                var i = 1;
                for (variable of stringVariables) {
                  var regex = new RegExp(variable + "(?![a-z]|_)", "g");
                  script.contents = script.contents.replace(regex, "&q" + i);
                  i++;
                }

                // Replace all " == " and " = " with "="
                script.contents = script.contents.replace(/ == /g, "=");
                script.contents = script.contents.replace(/ = /g, "=");

                // Replace all " + " with "+", similar for - * /
                script.contents = script.contents.replace(/ \+ /g, "+");
                script.contents = script.contents.replace(/ - /g, "-");
                script.contents = script.contents.replace(/ \* /g, "*");
                script.contents = script.contents.replace(/ \/ /g, "/");

                // Replace all " && " with "&&", similar for ||
                script.contents = script.contents.replace(/ && /g, "&&");
                script.contents = script.contents.replace(/ \|\| /g, "||");

                // Replace all " >= ", " > ", etc.
                script.contents = script.contents.replace(/ >= /g, ">=");
                script.contents = script.contents.replace(/ > /g, ">");
                script.contents = script.contents.replace(/ <= /g, "<=");
                script.contents = script.contents.replace(/ < /g, "<");
              }

              // If option is set, format the code as a block up to a certain width
              if (config.blockFormat) {
                var lines = script.contents.split(/[\r\n]+/);
                var block = "";
                var length = 0;
                for (line of lines) {
                  if (line.length + length <= config.maxWidth) {
                    block += line + ";";
                  } else {
                    block += "\r\n" + line + ";";
                    length = 0;
                  }
                  length += line.length;
                }
                script.contents = block;
              }

              // Add timestamp
              script.contents =
                "// " +
                moment().format("MMMM Do YYYY") +
                "\r\n" +
                script.contents;

              // Add personal message
              script.contents = config.message + "\r\n" + script.contents;

              // Write compiled macromod file
              fs.writeFile(
                config.scriptsFolder + script.compiledName,
                script.contents,
                err => {
                  if (err) {
                    throw err;
                  }

                  console.log(
                    moment().format("LTS") +
                      ": Recompiled " +
                      script.compiledName
                  );

                  // Set watch(es) to look for file changes, if not ommited
                  if (!argv.i) {
                    paths = [];
                    for (dependent of script.allDependents) {
                      paths.push(dependent.path);
                    }

                    paths = _.uniq(paths);
                    paths.push(script.path);

                    script.watcher = watch(paths, (evt, name) => {
                      recompile(evt, name, script);
                    });
                  }
                }
              );
            }
          }
        );
      };

      loopUntilNoDependents(script);
    });
  };

  recompile(0, 0, script);
}
